# $Id: quick_rank.gap,v 1.13 2002/06/26 20:24:25 nathand Exp $
#
# Comments:  For smallish matrices, the performance of RankMatModP
# is hampered by the fact that NPrintMatrix is remarkably slow, 
# compared to PrintTo(file, Matrix).  One could increase the speed 
# greatly by re-writing my_mat/quick_rank.c to accept that form of
# input.
#
# Also, one could reduce the (already small) number of false positives
# (covers with finite field betti number > 0 but with rational betti
# number equal to 0) by using a randomly choosen prime of size about
# 2^15, rather that always using 31991.

NumTimes := 0;
NQuickMatrixTmpFile := DirectoryTemporary();

QuickMatrixCommand := "my_mat/quick_rank";

NPrintMatrix := function(M, p, outstream)
  local i, j, dim;
  if IsMatrix( M ) and IsInt( M[1][1] ) then
    dim := DimensionsMat(M);
    AppendTo(outstream, p, " ", dim[1], " ", dim[2], "\n");
    for i in [ 1 .. dim[1] ] do
      for j in [ 1 .. dim[2] ] do
        AppendTo(outstream, M[i][j], " " );
      od;
      AppendTo(outstream, "\n");
    od;
  else
    return fail;
  fi;
end;

RankMatModP := function(M, p)  
  local a, dims, str, tmp, tmpfile, tmpoutfile, o;
  NumTimes := NumTimes + 1;
  if p >= 2^15 then 
    Error("prime p to large");
  fi;
  if not IsPrime(p) then
    Error("modulus not prime");
  fi;
  dims := DimensionsMat(M);
  if dims[1] = dims[2] and dims[1] <= 10 then 
    return(dims[1] - Number(ElementaryDivisorsMat(M), x -> x = 0));
  fi;
  if Maximum(dims) <= 10 then
     return RankMat(M);
  fi; 
  tmp := NQuickMatrixTmpFile;
  tmpfile :=  Filename(tmp, "mat");
  tmpoutfile := Filename(tmp, "out");
  o := OutputTextFile(tmpfile, false);
  NPrintMatrix(M,p,o);
  CloseStream(o);
  Exec(QuickMatrixCommand, tmpfile, "> ", tmpoutfile);
  RemoveFile(tmpfile);
  a := ReadAsFunction(tmpoutfile);
  RemoveFile(tmpoutfile);
  return a();
end;

RankMatModPExpectingFull := function(M, p)
   local tmp, z, maxrank;
   maxrank := Minimum(DimensionsMat(M));
   tmp := M * Z(251); 
   for z in tmp do
     CONV_VEC8BIT(z, 251); 
   od;
   if RankMat(tmp) = maxrank then 
     #Print("Good\n");
     return maxrank;
   else
     #Print("Bad\n");
     return RankMatModP(M, p);
   fi;
end;

# Note: AbelianRankOfSubgroup is really just computing mod 31991,
# dispite the name.

AbelianRankOfSubgroup := function(G, H)
	local M;
	M:=RelatorMatrixAbelianizedSubgroupRrs(G, H);
	return DimensionsMat(M)[2] - RankMatModP(M, 31991);
end;

AbelianInvariantRankOfSubgroup := function(G, H)
    local I;
    I := AbelianInvariantsSubgroupFpGroup(G, H);
    return Number(I, x -> x = 0);
end;

# Functions for checking the rank of a matrix over Z using
# "rational reconstruction"

RationalReconstruction := function(s, p)
	local u_curr, u_last, v_curr, v_last,temp,q, f, g, i;
	Assert(-1, IsPrimePowerInt(p));
	Assert(-1, 0 <= s and s < p);
	
	u_last := p;
	u_curr := s;
	v_last := 0;
	v_curr := 1;
	
	i := 0;
	while u_curr^2 >= p do
		q := QuoInt(u_last, u_curr);
	    
	    temp := u_last - q*u_curr;
	    u_last := u_curr;
	    u_curr := temp;
	    
	    temp := v_last + q*v_curr;
	    v_last := v_curr;
	    v_curr := temp;
	    
	    i := i + 1;
	od;
	f := (-1)^i * u_curr;
	g := v_curr;
	if 2*f^2 < p and 2*g^2 < p then
		return f/g;
	else
		return fail;
	fi;
end;

RankOfKernelOfIntegerMatrixViaRatRecon := function(M, p)
    local tmp, tmpfile,tmpoutfile, o, a, ker, row, lifted_row;
    # compute the kernal mod p
    
    tmp := NQuickMatrixTmpFile;
    tmpfile :=  Filename(tmp, "mat");
    tmpoutfile := Filename(tmp, "out");
    o := OutputTextFile(tmpfile, false);
    NPrintMatrix(M,p,o);
    CloseStream(o);
    Exec("NTL/rank", " < " , tmpfile, " > ", tmpoutfile);
    RemoveFile(tmpfile);
    a := ReadAsFunction(tmpoutfile);
    RemoveFile(tmpoutfile);
    ker := a();
    
    # now try to lift each row;
    
    for row in ker do
        lifted_row := List(row, x -> RationalReconstruction(x,p));
        if fail in lifted_row then
            Print("Row didn't lift.\n");            
            return fail;
        fi;
        
        if not IsZero( M * lifted_row ) then
            Print("Lifted row not in ker!\n");
            return fail;
        fi;
    od;
    return Length(ker);

end;
    
TrueAbelianRankOfSubgroup := function(G, H)
    local M, R;
    M:=RelatorMatrixAbelianizedSubgroupRrs(G, H);
    #R := RankOfKernelOfIntegerMatrixViaRatRecon(M, 100003);
    R := RankOfKernelOfIntegerMatrixViaRatRecon(M, 5000011);
    #R := RankOfKernelOfIntegerMatrixViaRatRecon(M, 5983742341223212234327);
    if R = fail then return 0; fi;
    return R; 
end;

#  Info for rank computing with Magma.
#  First get "RelatorMatrixAbelianizedSubgroupRrs(G, PreImage(f,U));"
#  by editing the appropriate line of "rigour.gap".  Then do
#  gap> A := CheckHomologyWithBlah(M, false, false);;
#  gap> PrintTo("mat.mag", A);
#   
#  Now edit "mat.mag" so it looks like 
#
#   X := Matrix(IntegerRing(),[[ ..data .. ]]);
#
#  Then start Magma
#
# >load "mat.mag";                 
# >K := Basis(NullspaceOfTranspose(X));
# >#K                        /* the rank of the homology */
# > IsZero(X * Transpose(Matrix(K)));    /* double check */

# Computes the number of bits worth of primes needed to check the
# Z-rank of the integer matrix M by computing the rank mod a
# collection of primes whose product has at least the returned number
# of bits.  Uses a Hadamard bound to estimate the size of a
# determinant of any sub matrix of M.

Hadamard := function(M)
       local R;
       R := List(M, r -> Sum(r, x->x^2));
       Sort(R);
       Reversed(R);
       return QuoInt(LogInt(Product( R{[1..DimensionsMat(M)[2] ]} ), 2),2);
end;

# Example of a manifold where the F_31991-betti number is positive but the
# Z-betti number is 0 is the A6-cover of "s926(-4,3)"


